/*
 * SSD1331にイメージデータを表示
 * 
 */
#include <SPI.h>
#include "PinAssign.h"
#include "OLED1331.h"
#include "font8x16.h"

SPIClass SPI_OLED ;

uint8_t SPIBuf[256] ;   // SPI転送用バッファ

// --- SSD1331 COMMADN ---
static uint8_t oledInitCommands[] = { // SSD1331の初期化コマンド群
  0xAE,         // DISPLAYOFF
  0xA0 , 0x72,  // SETREMAP 0x72 / 0x76 : RGB Color
// ▼ 設定を変更してみる場合の設定値
//  0xA0 , B00110010 ,  // ノーマル 256色
//  0xA0 , B01110010 ,  // ノーマル 64色 Format1
//  0xA0 , B10110010 ,  // ノーマル 64色 Format2
//  0xA0 , B01100010 ,  // 縦方向ミラー反転 64色 Format1
//  0xA0 , B01110000 ,  // 横方向ミラー反転 64色 Format1
//  0xA0 , B01100000 ,  // 上下反転　      64色 Format1
//  0xA0 , B01100011 ,  // 左90度回転表示　64色 Format1 // ※90度回転は何故か正常に動作しない
//  0xA0 , B01110001 ,  // 右90度回転表示　64色 Format1
  0xA1 , 0x00 , // STARTLINE
  0xA2 , 0x00 , // DISPLAYOFFSET
  0xA4 ,        // NORMALDISPLAY
  0xB0 , 0x0B , // POWERMODE
  0x87 , 0x01 , // MASTERCURRENT : 0x06
  0x81 , 0x91 , // CONTRAST RED  : 0x91
  0x82 , 0x50 , // CONTRAST GREEN: 0x50
  0x83 , 0x7D , // CONTRAST BLUR : 0x7D
  0xAF ,        // DISPLAYON
};

uint8_t drowComandClearWindow[] = {0x25 , 0 , 0 , 95, 63 }; // ClearWindow Command
uint8_t drowComandLine[] = {0x21 ,  0,  0,  0,  0, 0x00, 0x00, 0x00 }; // DrowLine Command
uint8_t drowComandRect[] = {0x22 , 0, 0, 95, 63, 0x3E, 0x3E, 0x3E, 0x3E, 0x3E, 0x3E }; // DrowRect Command

// ===== SPI WRITE =====
void OLED1331::init( ) {
    // -----  ST3375 INITIAL -----
    pinMode(PIN_DC, OUTPUT);
    pinMode(PIN_RESET, OUTPUT);
    pinMode(PIN_CS, OUTPUT);
    SPI_OLED.begin(PIN_CLK,PIN_MISO,PIN_MOSI,-1);  // SPI.begin(SCK, MISO, MOSI, SS);
    oldSelect( ) ;

    // --- Reset SSD1306
    pinMode(PIN_RESET, OUTPUT);
    digitalWrite(PIN_RESET, HIGH);
    delay(1);                       // VDD goes high at start, pause for 1 ms
    digitalWrite(PIN_RESET, LOW);  // Bring reset low
    delay(10);                      // Wait 10 ms
    digitalWrite(PIN_RESET, HIGH); // Bring out of reset

    // --- SSD1306 初期化
    // OLED設定コマンド送信
    oledSendCommands(sizeof(oledInitCommands), oledInitCommands) ;  // 初期化
    oledSendCommand(0xA4) ; // 描画ON
    oledSendCommand(0xAF) ; // 画面表示

    // ----- CLEAR WINDOW -----
    memcpy(SPIBuf, drowComandClearWindow, sizeof(drowComandClearWindow)) ;
    oledSendCommands(sizeof(drowComandClearWindow), SPIBuf) ;  // ClearWindow Command
    delayMicroseconds(500);
}

void OLED1331::oldSelect( ) {
    digitalWrite(PIN_CS, HIGH);
}

// OLEDに1バイトコマンドを送信
void OLED1331::oledSendCommand(uint8_t command) {
    digitalWrite(PIN_DC, LOW); // Command mode
    SPI_OLED.transfer(command);
}

// OLEDに2バイトコマンドを送信
void OLED1331::oledSendCommand2(uint8_t command1, uint8_t command2) {
    digitalWrite(PIN_DC, LOW); // Command mode
    SPI_OLED.transfer(command1);
    SPI_OLED.transfer(command2);
}

// OLEDに3バイトコマンドを送信
void OLED1331::oledSendCommand3(uint8_t command1, uint8_t command2, uint8_t command3) {
    digitalWrite(PIN_DC, LOW); // Command mode
    SPI_OLED.transfer(command1);
    SPI_OLED.transfer(command2);
    SPI_OLED.transfer(command3);
}

// OLEDに複数のコマンドを送信
void OLED1331::oledSendCommands(int commandLen, uint8_t commands[]) {
    digitalWrite(PIN_DC, LOW); // Command mode
    SPI_OLED.transfer(commands, commandLen);
}

void OLED1331::drowBitMap(int offsetX , int offsetY , int width, int hight, uint8_t bitMap[]) {
    int dispWide = (width > OLED_SSD1331_WIDTH) ? OLED_SSD1331_WIDTH : width ;
    int dispHight= (hight > OLED_SSD1331_HIGHT) ? OLED_SSD1331_HIGHT : hight ;
    uint8_t endX = offsetX + dispWide - 1;
    uint8_t endY = offsetY + dispHight - 1;
    oledSendCommand3(0x15, offsetX, endX) ;
    oledSendCommand3(0x75, offsetY, endY) ;
    delayMicroseconds(500);

    int pos = 0 ;
    offsetX = (dispWide - OLED_SSD1331_WIDTH) / 2 ;
    offsetX = (offsetX < 0) ? 0 : offsetX ;
    offsetY = (dispHight - OLED_SSD1331_HIGHT) / 2 ;
    offsetY = (offsetY < 0) ? 0 : offsetY ;
    digitalWrite(PIN_DC, HIGH); // Data mode
    for (int row = 0 ; row < dispHight ; row ++) {
        pos = (row+offsetY) * width * 2 + offsetX * 2 ;
      memcpy(SPIBuf,&bitMap[pos],dispWide * 2) ;
      SPI_OLED.transfer(SPIBuf, dispWide * 2);
    }
    digitalWrite(PIN_DC, LOW); // Command mode
}

void OLED1331::drowBitMap(IMAGEDATA *imageData){
    int width  = imageData->getWidth();
    int hight = imageData->getHight() ;
    int dispWide = (width > OLED_SSD1331_WIDTH) ? OLED_SSD1331_WIDTH : width ;
    int dispHight= (hight > OLED_SSD1331_HIGHT) ? OLED_SSD1331_HIGHT : hight ;
    uint8_t offsetX = (width > OLED_SSD1331_WIDTH) ? (width > OLED_SSD1331_WIDTH) / 2 : 0 ;
    uint8_t offsetY = (hight > OLED_SSD1331_HIGHT) ? (hight > OLED_SSD1331_HIGHT) / 2 : 0 ;

    uint8_t endX = dispWide - 1;
    uint8_t endY = dispHight - 1;
    oledSendCommand3(0x15, 0, endX) ;
    oledSendCommand3(0x75, 0, endY) ;
    delayMicroseconds(500);

    int pos = 0 ;
    uint8_t *bitMap = imageData->getImage( ) ;
    digitalWrite(PIN_DC, HIGH); // Data mode
    for (int row = 0 ; row < dispHight ; row ++) {
        pos = (row+offsetY) * width * 2 + offsetX * 2 ;
        memcpy(SPIBuf,&bitMap[pos],dispWide * 2) ;
        SPI_OLED.transfer(SPIBuf, dispWide * 2);
    }
    digitalWrite(PIN_DC, LOW); // Command mode
}

void OLED1331::drowLine(int xs,int ys,int xe,int ye,uint32_t color) {
    memcpy(SPIBuf, drowComandLine, sizeof(drowComandLine)) ;
    SPIBuf[1] = xs ; SPIBuf[2] = ys ;
    SPIBuf[3] = xe ; SPIBuf[4] = ye ;
    SPIBuf[5] =(color & 0x00ff0000) >> 16 ;
    SPIBuf[6] =(color & 0x0000ff00) >>  8 ;
    SPIBuf[7] =color & 0x000000ff ;
    oledSendCommands(sizeof(drowComandLine), SPIBuf) ;  // DrowLine Command
}

void OLED1331::cls( ) {
    // ----- CLEAR WINDOW -----
    memcpy(SPIBuf, drowComandClearWindow, sizeof(drowComandClearWindow)) ;
    oledSendCommands(sizeof(drowComandClearWindow), SPIBuf) ;  // ClearWindow Command
    delayMicroseconds(500);
}
void OLED1331::cls(uint32_t color) {
    // ----- RECT -----
    oledSendCommand2(0x26, 0x01) ; // 塗りつぶしON
    delayMicroseconds(500);
    memcpy(SPIBuf, drowComandRect, sizeof(drowComandRect)) ;
    SPIBuf[5] = SPIBuf[8] = (color & 0x00ff0000) >> 16 ;
    SPIBuf[6] = SPIBuf[9] = (color & 0x0000ff00) >>  8 ;
    SPIBuf[7] = SPIBuf[10] = color & 0x000000ff ;
    oledSendCommands(sizeof(drowComandRect), SPIBuf) ;  // DrowLine Command
    delayMicroseconds(500);
}

void OLED1331::drowStr(char str[] , uint8_t x , uint8_t y,uint16_t fontColor , uint16_t bgColor) {
    for (int i=0;str[i] != 0 ; i++) {
        drowFont(str[i],x,y,fontColor,bgColor) ;
        x += 8 ;
        if (x+8>=96) {
          x = 0 ;
          y += 16 ;
          if (y + 16 >= 64) {
            y = 0 ; 
          }
        }
    }
}

void OLED1331::drowFont(uint8_t code , uint8_t x, uint8_t y,uint16_t fontColor , uint16_t bgColor) {
    int pos = (code - 0x20) * 16 ;
    uint8_t fontColor0 = (fontColor & 0xff00) >> 8 ;
    uint8_t fontColor1 = fontColor & 0x00ff ;
    uint8_t bgColor0 = (bgColor & 0xff00) >> 8 ;
    uint8_t bgColor1 = bgColor & 0x00ff ;
    
    memset(SPIBuf,0,8*16*2) ;
    for (int i=0;i<16;i++) {
      uint8_t data = ssd1306xled_font8x16[pos + i];
      uint8_t bitMast = 0x01 ; 
      for (int j=0;j<8;j++) {
        int fpos = 0 ;
        if (i<8) {
          fpos = (j*8+i)*2 ;
        } else {
          fpos = ((j+8)*8+(i-8))*2 ;
        }
        if ((data & bitMast) == 0) {
          SPIBuf[fpos] = bgColor0 ;
          SPIBuf[fpos+1] = bgColor1 ;
        } else {
          SPIBuf[fpos] = fontColor0 ;
          SPIBuf[fpos+1] = fontColor1 ;
        }
        bitMast <<= 1 ;
      }
    }
  
    uint8_t endX = x + 7 ;
    uint8_t endY = y + 15;
    oledSendCommand3(0x15, x, endX) ;
    oledSendCommand3(0x75, y, endY) ;
    delayMicroseconds(500);
    digitalWrite(PIN_DC, HIGH); // Data mode
    SPI_OLED.transfer(SPIBuf, 8*16*2);
    digitalWrite(PIN_DC, LOW); // Command mode
    delayMicroseconds(500);
}
