#include <SPI.h>
#include "PinAssign.h"
#include "SDCARD.h"
SPIClass SPI_SD ;
void SDCARD::init( ){
#if DEBUG
  Serial.println("initialization Start");
#endif
  pinMode(PIN_CS, OUTPUT);
  digitalWrite(PIN_CS, LOW);
  SPI_SD.begin(PIN_CLK,PIN_MISO,PIN_MOSI);  // SPI.begin(SCK, MISO, MOSI, SS);
  if (!SD.begin(PIN_CS,SPI_SD,4000000)) {
    Serial.println("initialization failed!");
    while (1);
  }
#if DEBUG
  Serial.println("initialization done.");
#endif
}

void SDCARD::sdCardSelect( ) {
  digitalWrite(PIN_CS, LOW);
}

void SDCARD::fileList( ) {
    Serial.println("fileList");
    File root = SD.open("/");
    printDirectory(root, 0);
    Serial.println("done!");
}

void SDCARD::printDirectory(File dir, int numTabs) {

    Serial.print("dir:");
    Serial.print(dir);
    Serial.print(" / numTabs=");
    Serial.println(numTabs);
    
    while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

UINT in_func (JDEC* jd, BYTE* buff, UINT nbyte)
{
  UINT ret = 0;
  IODEV *dev = (IODEV*)jd->device;   // Device identifier for the session (5th argument of jd_prepare function) 
  File *f2 = (File *)dev->fp;
  if (buff) {
    return (UINT)f2->read(buff, nbyte);
  } else {
    // Remove bytes from input stream 
    if(f2->seek(nbyte, SeekCur)){ //seek mode:SeekCur=1(FS.h)
      ret = nbyte;
    }else{
      ret = 0;
    }
  }
  return ret;
}
 
UINT out_func (JDEC* jd, void* bitmap, JRECT* rect)
{
  IODEV *dev = (IODEV*)jd->device;
  uint8_t *src, *dst;
  uint16_t x, y, bwd;
 
#if DEBUG
  if (rect->left == 0) {
    Serial.printf("  %lu%%\r\n", (rect->top << jd->scale) * 100UL / jd->height);
  }
#endif

  src = (uint8_t*)bitmap;
  //use RGB565
  dst = dev->fbuf + 2 * (rect->top * dev->wfbuf + rect->left);
  //bws = 2 * (rect->right - rect->left + 1);
  bwd = 2 * dev->wfbuf;
  uint8_t msb, lsb, tmp1, tmp2;
  int16_t xcnt;
  for (y = rect->top; y <= rect->bottom; y++) {
    xcnt = 0;
    for (x = rect->left; x <= rect->right; x++){
      msb = *(src++) & 0b11111000;
      tmp1 = *(src++) & 0b11111100;
      msb = msb | (tmp1 >> 5);
      tmp2 = *(src++) & 0b11111000;
      lsb = (tmp1 << 3) | (tmp2 >> 3);
      *(dst + xcnt++) = msb;
      *(dst + xcnt++) = lsb;
    }
    dst += bwd;
  }
 
  return 1;    // Continue to decompress 
}

bool SDCARD::LoadImage(char *fileName,IMAGEDATA *imageData) {
  void *work;       /* Pointer to the decompressor work area */
  JDEC jdec;        /* Decompression object */
  JRESULT res;      /* Result code of TJpgDec API */
  IODEV devid;      /* User defined device identifier */

  bool result = true ;

  File file = SD.open(fileName);
  if(!file){
    Serial.println("Failed to open file for reading");
    return false;
  }
#if DEBUG
  Serial.printf("%s file size = %d\r\n", fileName, file.size());
  Serial.printf("1.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif
  devid.fp = &file;
 
  // Allocate a work area for TJpgDec 
  const int WorkBuffSize = 3100 ;
  work = malloc(WorkBuffSize);
  // Prepare to decompress 
  res = jd_prepare(&jdec, in_func, work, WorkBuffSize, &devid);
  if (res == JDR_OK) {
    // Ready to dcompress. Image info is available here. 
#if DEBUG
    Serial.printf("Image dimensions: %u by %u. %u bytes used.\r\n", jdec.width, jdec.height, WorkBuffSize - jdec.sz_pool);
#endif
    imageData->setWidth(jdec.width) ;
    imageData->setHight(jdec.height) ;
    imageData->allocImage(2 * jdec.width * jdec.height) ; //RGB565
    devid.fbuf = (BYTE *)imageData->getImage( ) ;
    if (devid.fbuf != NULL) {
      devid.wfbuf = jdec.width;
  
      res = jd_decomp(&jdec, out_func, 0);   // Start to decompress with 1/1 scaling 
#if DEBUG
      if (res == JDR_OK) {
        // Decompression succeeded. You have the decompressed image in the frame buffer here. 
        Serial.printf("\rOK  \n");
      } else {
        Serial.printf("Failed to decompress: rc=%d\n", res);
      }
#endif
    } else {
      Serial.printf("Out of Memory\r\n");
      result = false ;
    }
  } else {
    Serial.printf("Failed to prepare: rc=%d\n", res);
    result = false ;
  }
  free(work);             // Discard work area 
  work = NULL ;

  if (!result) {
    return result ;
  }
 
#if DEBUG
  Serial.printf("2.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif
//  delay(500);
  if(devid.fbuf){
    devid.fbuf = NULL;
#if DEBUG
    Serial.println("Free devid.fbuf OK!");
  }else{
    Serial.println("Failed free devid.fbuf.");
#endif
  }

#if DEBUG
  Serial.printf("3.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif
  return result ;
}
